/*
 * PmodACL.c
 *
 *  Created on: Oct 31, 2017
 *      Author: Krzysztof Chojnowski
 */
#include "stm32l4xx_hal.h"

#include <PmodACL.h>

#define MAX_SPI_RX_BUFER_LEN 32 // Maximum number of bytes that can be read from SPI

// The read/write and MB flags have to be added to SPI transaction address byte.
#define SPI_READ_FLAG   0x80
#define SPI_WRITE_FLAG  0x00
#define SPI_MB_FLAG     0x40

SPI_HandleTypeDef pmodAclSpi;

static void writeRegister(uint8_t address, uint8_t data)
{
  uint8_t txbuf[2] = {address | SPI_WRITE_FLAG | SPI_MB_FLAG, data};

  HAL_GPIO_WritePin(GPIOB, GPIO_PIN_0, GPIO_PIN_RESET);
  HAL_SPI_Transmit(&pmodAclSpi, txbuf, 2, 100);
  HAL_GPIO_WritePin(GPIOB, GPIO_PIN_0, GPIO_PIN_SET);
}

static void readRegister(uint8_t address, uint8_t* data, uint8_t size)
{
  // Reading N bytes requires one more byte at the beginning for address transmission,
  // so the whole SPI transaction has to be N + 1 bytes long.
  uint8_t txbuf[MAX_SPI_RX_BUFER_LEN + 1] = {0x00};
  uint8_t rxbuf[MAX_SPI_RX_BUFER_LEN + 1] = {0x00};
  txbuf[0] = address | SPI_READ_FLAG | SPI_MB_FLAG;

  HAL_GPIO_WritePin(GPIOB, GPIO_PIN_0, GPIO_PIN_RESET);
  HAL_SPI_TransmitReceive(&pmodAclSpi, txbuf, rxbuf, size + 1, 100);
  HAL_GPIO_WritePin(GPIOB, GPIO_PIN_0, GPIO_PIN_SET);

  for(int i=0; i<size; i++)
    data[i] = rxbuf[i+1];
}

void PmodACL_Config(void)
{
  // Enable clock for all required GPIO ports and SPI1.
  __HAL_RCC_SPI1_CLK_ENABLE();
  __HAL_RCC_GPIOA_CLK_ENABLE();
  __HAL_RCC_GPIOB_CLK_ENABLE();
  __HAL_RCC_GPIOE_CLK_ENABLE();

  // Configure the SPI connected to the Pmod module.
  pmodAclSpi.Instance = SPI1;
  pmodAclSpi.Init.Mode = SPI_MODE_MASTER;
  pmodAclSpi.Init.Direction = SPI_DIRECTION_2LINES;
  pmodAclSpi.Init.DataSize = SPI_DATASIZE_8BIT;
  pmodAclSpi.Init.CLKPolarity = SPI_POLARITY_HIGH;
  pmodAclSpi.Init.CLKPhase = SPI_PHASE_2EDGE;
  pmodAclSpi.Init.NSS = SPI_NSS_SOFT;
  pmodAclSpi.Init.BaudRatePrescaler = SPI_BAUDRATEPRESCALER_128;
  pmodAclSpi.Init.FirstBit = SPI_FIRSTBIT_MSB;
  pmodAclSpi.Init.TIMode = SPI_TIMODE_DISABLE;
  pmodAclSpi.Init.CRCCalculation = SPI_CRCCALCULATION_DISABLE;
  pmodAclSpi.Init.NSSPMode = SPI_NSS_PULSE_DISABLE;

  HAL_SPI_Init(&pmodAclSpi);

  // Initialization of PmodACL.
  writeRegister(0x1d, 0x3C); // Set the tap threshold (THRESH_TAP register) to 60 * 62.5mg.
  writeRegister(0x21, 0x0A); // Set the tap duration (DUR register) to 10 * 625us.
  writeRegister(0x22, 0x10); // Set the latent (Latent register) to 16 * 1.25ms.
  writeRegister(0x23, 0x80); // Set the window duration (Window register) to 128 * 1.25m.
  writeRegister(0x2a, 0x07); // Enable all axes for tap detection (TAP_AXES register).
  writeRegister(0x2d, 0x08); // Enable measurement in POWER_CTL register.

  // Clear all flags before enabling the interrupt.
  PmodACL_ReadInterruptFlags();

  writeRegister(0x2e, 0x20); // Enable double tab interrupt in INT_ENABLE register.
  writeRegister(0x31, 0x28); // Set the interrupts to be active in low state and enable
                             // full resolution in DATA_FORMAT register.

  // Configure the INT1 (PE13) interrupt line with the lowest priority.
  GPIO_InitTypeDef GPIO_InitStruct;
  GPIO_InitStruct.Mode = GPIO_MODE_IT_FALLING;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
  GPIO_InitStruct.Pin = GPIO_PIN_13;
  HAL_GPIO_Init(GPIOE, &GPIO_InitStruct);

  HAL_NVIC_SetPriority(EXTI15_10_IRQn, 0x0F, 0);
  HAL_NVIC_EnableIRQ(EXTI15_10_IRQn);
}

void HAL_SPI_MspInit(SPI_HandleTypeDef *hspi)
{
  // Initialize GPIO used by the SPI1 peripheral. The CS is controlled by the software.
  GPIO_InitTypeDef GPIO_InitStruct;

  GPIO_InitStruct.Mode = GPIO_MODE_AF_PP;
  GPIO_InitStruct.Pull = GPIO_PULLDOWN;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
  GPIO_InitStruct.Alternate = GPIO_AF5_SPI2;
  GPIO_InitStruct.Pin = GPIO_PIN_1 | GPIO_PIN_7;
  HAL_GPIO_Init(GPIOA, &GPIO_InitStruct);

  GPIO_InitStruct.Pin = GPIO_PIN_14;
  HAL_GPIO_Init(GPIOE, &GPIO_InitStruct);

  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_PULLDOWN;
  GPIO_InitStruct.Pin = GPIO_PIN_0;
  HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);
  HAL_GPIO_WritePin(GPIOB, GPIO_PIN_0, GPIO_PIN_SET);
}

void PmodACL_ReadXYZ(int16_t* x, int16_t* y, int16_t* z)
{
  uint8_t data[6];
  readRegister(0x32, data, 6);

  // The ADXL345 on PmodACL is configured for the highest resolution (4mg/LSB)
  // so the axes readings need to be multiplied by 4 to obtain the acceleration in [mg].
  *x = ((data[1] << 8) + data[0]) * 4;
  *y = ((data[3] << 8) + data[2]) * 4;
  *z = ((data[5] << 8) + data[4]) * 4;
}

uint8_t PmodACL_ReadInterruptFlags(void)
{
  uint8_t data;
  readRegister(0x30, &data, 1);
  return data;
}
